from cached_property import cached_property
from csv import writer
from onegov.core.collection import Pagination
from onegov.swissvotes.fields.dataset import COLUMNS
from onegov.swissvotes.fields.dataset import DatasetDialect
from onegov.swissvotes.models import PolicyArea
from onegov.swissvotes.models import SwissVote
from sqlalchemy import or_


class SwissVoteCollection(Pagination):

    batch_size = 20
    initial_sort_by = 'date'
    initial_sort_order = 'descending'
    default_sort_order = 'ascending'

    SORT_BYS = ('date', 'legal_form', 'result', 'result_people_yeas_p',
                'title')
    SORT_ORDERS = ('ascending', 'descending')

    def __init__(
        self,
        session,
        page=None,
        from_date=None,
        to_date=None,
        legal_form=None,
        result=None,
        policy_area=None,
        sort_by=None,
        sort_order=None
    ):
        self.session = session
        self.page = page
        self.from_date = from_date
        self.to_date = to_date
        self.legal_form = legal_form
        self.result = result
        self.policy_area = policy_area
        self.sort_by = sort_by
        self.sort_order = sort_order

    def subset(self):
        return self.query()

    def __eq__(self, other):
        return (
            (self.page or 0) == (other.page or 0) and
            (self.from_date or None) == (other.from_date or None) and
            (self.to_date or None) == (other.to_date or None) and
            set(self.legal_form or []) == set(other.legal_form or []) and
            set(self.result or []) == set(other.result or []) and
            set(self.policy_area or []) == set(other.policy_area or []) and
            (self.sort_by or None) == (other.sort_by or None) and
            (self.sort_order or None) == (other.sort_order or None)
        )

    @property
    def page_index(self):
        """ The current page. """

        return self.page or 0

    def page_by_index(self, page):
        """ Returns the requested page. """

        return self.__class__(
            self.session,
            page=page,
            from_date=self.from_date,
            to_date=self.to_date,
            legal_form=self.legal_form,
            result=self.result,
            policy_area=self.policy_area,
            sort_by=self.sort_by,
            sort_order=self.sort_order
        )

    @property
    def current_sort_by(self):
        """ Returns the currently used sorting key.

        Defaults to a reasonable value.

        """
        result = self.initial_sort_by if self.sort_by is None else self.sort_by
        return result if result in self.SORT_BYS else self.initial_sort_by

    @property
    def current_sort_order(self):
        """ Returns the currently used sorting order.

        Defaults to a reasonable value.

        """

        default = self.default_sort_order
        default = self.initial_sort_order if self.sort_by is None else default
        result = default if self.sort_order is None else self.sort_order
        return result if result in self.SORT_ORDERS else default

    def sort_order_by_key(self, sort_by):
        """ Returns the sort order by key.

        Defaults to 'unsorted'.

        """

        if self.current_sort_by == sort_by:
            return self.current_sort_order
        return 'unsorted'

    def by_order(self, sort_by):
        """ Returns the votes ordered by the given key. """

        sort_order = self.default_sort_order
        if sort_by == self.current_sort_by:
            if self.current_sort_order == 'ascending':
                sort_order = 'descending'
            else:
                sort_order = 'ascending'

        return self.__class__(
            self.session,
            page=None,
            from_date=self.from_date,
            to_date=self.to_date,
            legal_form=self.legal_form,
            result=self.result,
            policy_area=self.policy_area,
            sort_by=sort_by,
            sort_order=sort_order
        )

    @property
    def order_by(self):
        """ Returns an SqlAlchemy expression for ordering queries based
        on the current sorting key and ordering.

        """

        if self.current_sort_by == 'title':
            from onegov.core.orm.func import unaccent
            result = unaccent(SwissVote.title)
        else:
            result = (
                getattr(SwissVote, f'_{self.current_sort_by}', None) or
                getattr(SwissVote, self.current_sort_by, None)
            )
            if not result:
                raise NotImplementedError()

        if self.current_sort_order == 'descending':
            result = result.desc()

        return result

    @property
    def offset(self):
        """ The current position in the batch. """

        return (self.page or 0) * self.batch_size

    @property
    def previous(self):
        """ The previous page. """

        if (self.page or 0) - 1 >= 0:
            return self.page_by_index((self.page or 0) - 1)

    @property
    def next(self):
        """ The next page. """

        if (self.page or 0) + 1 < self.pages_count:
            return self.page_by_index((self.page or 0) + 1)

    def query(self):
        """ Returns the votes matching to the current filters and order. """

        query = self.session.query(SwissVote)

        if self.from_date:
            query = query.filter(SwissVote.date >= self.from_date)
        if self.to_date:
            query = query.filter(SwissVote.date <= self.to_date)
        if self.legal_form:
            query = query.filter(SwissVote._legal_form.in_(self.legal_form))
        if self.result:
            query = query.filter(
                SwissVote._result.in_(self.result)
            )
        if self.policy_area:
            levels = [[], [], []]
            for area in self.policy_area:
                area = PolicyArea(area)
                if area.level == 1:
                    levels[0].append(area.descriptor_decimal)
                if area.level == 2:
                    levels[1].append(area.descriptor_decimal)
                if area.level == 3:
                    levels[2].append(area.descriptor_decimal)
            if levels[0]:
                query = query.filter(
                    or_(
                        SwissVote.descriptor_1_level_1.in_(levels[0]),
                        SwissVote.descriptor_2_level_1.in_(levels[0]),
                        SwissVote.descriptor_3_level_1.in_(levels[0])
                    )
                )
            if levels[1]:
                query = query.filter(
                    or_(
                        SwissVote.descriptor_1_level_2.in_(levels[1]),
                        SwissVote.descriptor_2_level_2.in_(levels[1]),
                        SwissVote.descriptor_3_level_2.in_(levels[1])
                    )
                )
            if levels[2]:
                query = query.filter(
                    or_(
                        SwissVote.descriptor_1_level_3.in_(levels[2]),
                        SwissVote.descriptor_2_level_3.in_(levels[2]),
                        SwissVote.descriptor_3_level_3.in_(levels[2])
                    )
                )

        query = query.order_by(self.order_by)

        return query

    def by_bfs_number(self, bfs_number):
        """ Returns the vote with the given BFS number. """

        query = self.query().filter(SwissVote.bfs_number == bfs_number)
        return query.first()

    @cached_property
    def available_descriptors(self):
        """ Returns a list of the used descriptor values (level 1-3). """

        # todo: optimize
        return[
            [
                x[0] for x in (
                    set(self.session.query(SwissVote.descriptor_1_level_1)) |
                    set(self.session.query(SwissVote.descriptor_2_level_1)) |
                    set(self.session.query(SwissVote.descriptor_2_level_1))
                ) if x[0]
            ],
            [
                x[0] for x in (
                    set(self.session.query(SwissVote.descriptor_1_level_2)) |
                    set(self.session.query(SwissVote.descriptor_2_level_2)) |
                    set(self.session.query(SwissVote.descriptor_2_level_2))
                ) if x[0]
            ],
            [
                x[0] for x in (
                    set(self.session.query(SwissVote.descriptor_1_level_3)) |
                    set(self.session.query(SwissVote.descriptor_2_level_3)) |
                    set(self.session.query(SwissVote.descriptor_2_level_3))
                ) if x[0]
            ]
        ]

    def update(self, votes):
        """ Adds or updates the given votes. """

        for vote in votes:
            old = self.query().filter_by(bfs_number=vote.bfs_number).first()
            if old:
                for attribute in COLUMNS.keys():
                    setattr(old, attribute, getattr(vote, attribute))
            else:
                self.session.add(vote)

    def export(self, file):
        """ Exports all votes according to the code book. """

        csv = writer(file, dialect=DatasetDialect())
        csv.writerow(COLUMNS.values())

        for vote in self.query().order_by(None).order_by(SwissVote.bfs_number):
            row = []
            for attribute, column in COLUMNS.items():
                value = getattr(vote, attribute)
                type_ = str(vote.__table__.columns[attribute.lstrip('_')].type)
                if value is None:
                    row.append('.')
                elif type_ == 'TEXT':
                    row.append(value)
                elif type_ == 'DATE':
                    row.append(f'{value:%d.%m.%Y}')
                elif type_ == 'INTEGER':
                    row.append(str(value))
                elif type_ == 'INT4RANGE':
                    row.append(f'{value.lower}-{value.upper}')
                elif type_.startswith('NUMERIC'):
                    row.append(
                        f'{value:f}'.replace('.', ',').rstrip('0').rstrip(',')
                    )
            csv.writerow(row)



from cached_property import cached_property
from onegov.core import Framework
from onegov.core import utils
from onegov.file import DepotApp
from onegov.form import FormApp
from onegov.quill import QuillApp
from onegov.swissvotes.models import Principal
from onegov.swissvotes.theme import SwissvotesTheme


class SwissvotesApp(Framework, FormApp, QuillApp, DepotApp):
    """ The swissvotes application. Include this in your onegov.yml to serve
    it with onegov-server.

    """

    serve_static_files = True

    def configure_sentry(self, **cfg):
        self.sentry_js = cfg.get('sentry_js')

    @cached_property
    def principal(self):
        return Principal()


@SwissvotesApp.static_directory()
def get_static_directory():
    return 'static'


@SwissvotesApp.template_directory()
def get_template_directory():
    return 'templates'


@SwissvotesApp.setting(section='core', name='theme')
def get_theme():
    return SwissvotesTheme()


@SwissvotesApp.setting(section='i18n', name='localedirs')
def get_i18n_localedirs():
    return [
        utils.module_path('onegov.swissvotes', 'locale'),
        utils.module_path('onegov.form', 'locale'),
        utils.module_path('onegov.user', 'locale')
    ]


@SwissvotesApp.setting(section='i18n', name='locales')
def get_i18n_used_locales():
    return {'de_CH', 'fr_CH'}


@SwissvotesApp.setting(section='i18n', name='default_locale')
def get_i18n_default_locale():
    return 'de_CH'


@SwissvotesApp.webasset_path()
def get_shared_assets_path():
    return utils.module_path('onegov.shared', 'assets/js')


@SwissvotesApp.webasset_path()
def get_js_path():
    return 'assets/js'


@SwissvotesApp.webasset_path()
def get_css_path():
    return 'assets/css'


@SwissvotesApp.webasset_output()
def get_webasset_output():
    return 'assets/bundles'


@SwissvotesApp.webasset('common')
def get_common_asset():
    yield 'modernizr.js'
    yield 'jquery.js'
    yield 'jquery.tablesorter.js'
    yield 'tablesaw.css'
    yield 'tablesaw.jquery.js'
    yield 'tablesaw-create.js'
    yield 'tablesaw-init.js'
    yield 'chosen.css'
    yield 'chosen.jquery.js'
    yield 'foundation.js'
    yield 'underscore.js'
    yield 'form_dependencies.js'
    yield 'common.js'



from onegov.core.converters import extended_date_converter
from onegov.core.i18n import SiteLocale
from onegov.swissvotes.app import SwissvotesApp
from onegov.swissvotes.collections import SwissVoteCollection
from onegov.swissvotes.collections import TranslatablePageCollection
from onegov.swissvotes.models import Principal
from onegov.swissvotes.models import SwissVote
from onegov.swissvotes.models import TranslatablePage
from onegov.user import Auth


@SwissvotesApp.path(
    model=Principal,
    path='/'
)
def get_principal(app):
    return app.principal


@SwissvotesApp.path(
    model=Auth,
    path='/auth'
)
def get_auth(request, to='/'):
    return Auth.from_request(request, to)


@SwissvotesApp.path(
    model=SiteLocale,
    path='/locale/{locale}'
)
def get_locale(request, app, locale, to=None):
    to = to or request.link(app.principal)
    return SiteLocale.for_path(app, locale, to)


@SwissvotesApp.path(
    model=SwissVoteCollection,
    path='/votes',
    converters=dict(
        page=int,
        from_date=extended_date_converter,
        to_date=extended_date_converter,
        legal_form=[int],
        result=[int],
        policy_area=[str],
        sort_by=str,
        sort_order=str
    )
)
def get_votes(
    app,
    page=None,
    from_date=None,
    to_date=None,
    legal_form=None,
    result=None,
    policy_area=None,
    sort_by=None,
    sort_order=None
):
    return SwissVoteCollection(
        app.session(),
        page=page,
        from_date=from_date,
        to_date=to_date,
        legal_form=legal_form,
        result=result,
        policy_area=policy_area,
        sort_by=sort_by,
        sort_order=sort_order
    )


@SwissvotesApp.path(
    model=SwissVote,
    path='/vote/{bfs_number}'
)
def get_vote(app, bfs_number):
    return SwissVoteCollection(app.session()).by_bfs_number(bfs_number)


@SwissvotesApp.path(
    model=TranslatablePage,
    path='/page/{id}'
)
def get_page(app, id):
    return TranslatablePageCollection(app.session()).by_id(id)



import click

from onegov.core.cli import abort
from onegov.core.cli import command_group
from onegov.core.cli import pass_group_context
from onegov.swissvotes import _
from onegov.swissvotes.models import TranslatablePage
from sqlalchemy import create_engine


cli = command_group()


@cli.command(context_settings={'creates_path': True})
@pass_group_context
def add(group_context):
    """ Adds an instance to the database. For example:

        onegov-swissvotes --select '/onegov_swissvotes/swissvotes' add

    """

    def add_instance(request, app):
        app.cache.invalidate()

        translators = {
            locale: app.translations.get(locale) for locale in app.locales
        }
        session = app.session()
        for page, title in (
            ('home', _("Homepage")),
            ('dataset', _("Dataset")),
            ('about', _("About")),
            ('contact', _("Contact")),
            ('disclaimer', _("Disclaimer")),
            ('imprint', _("Imprint")),
        ):
            translations = {
                locale: title.interpolate(translators[locale].gettext(title))
                for locale in app.locales
            }
            session.add(
                TranslatablePage(
                    id=page,
                    title_translations=translations,
                    content_translations=translations
                )
            )

        click.echo("Instance was created successfully")

    return add_instance


@cli.command()
@pass_group_context
def delete(group_context):
    """ Deletes an instance from the database. For example:

        onegov-swissvotes --select '/onegov_swissvotes/swissvotes' delete

    """

    def delete_instance(request, app):

        confirmation = "Do you really want to DELETE {}?".format(app.schema)

        if not click.confirm(confirmation):
            abort("Deletion process aborted")

        assert app.has_database_connection
        assert app.session_manager.is_valid_schema(app.schema)

        dsn = app.session_manager.dsn
        app.session_manager.session().close_all()
        app.session_manager.dispose()

        engine = create_engine(dsn)
        engine.execute('DROP SCHEMA "{}" CASCADE'.format(app.schema))
        engine.raw_connection().invalidate()
        engine.dispose()

        click.echo("Instance was deleted successfully")

    return delete_instance



from onegov.form import Form
from onegov.form.fields import MultiCheckboxField
from onegov.swissvotes import _
from onegov.swissvotes.collections import SwissVoteCollection
from onegov.swissvotes.fields import PolicyAreaField
from onegov.swissvotes.models import PolicyArea
from onegov.swissvotes.models.policy_area import POLICY_AREA
from onegov.swissvotes.models.vote import LEGAL_FORM
from onegov.swissvotes.models.vote import RESULT
from wtforms.fields.html5 import DateField
from wtforms import HiddenField


class SearchForm(Form):

    from_date = DateField(
        label=_("From date")
    )

    to_date = DateField(
        label=_("To date")
    )

    policy_area = PolicyAreaField(
        label=_("Policy area"),
        choices=[]
    )

    legal_form = MultiCheckboxField(
        label=_("Legal form"),
        choices=list(LEGAL_FORM.items()),
        coerce=int
    )

    result = MultiCheckboxField(
        label=_("Voting result"),
        choices=list(RESULT.items()),
        coerce=int
    )

    sort_by = HiddenField()
    sort_order = HiddenField()

    def populate_policy_area(self):
        votes = SwissVoteCollection(self.request.session)
        available = votes.available_descriptors

        def add_choice(value, label, level):
            self.policy_area.choices.append(
                (value, '‧' * level + ' ' + self.request.translate(label))
            )

        self.policy_area.choices = []
        for key_1, value_1 in POLICY_AREA.items():
            area = PolicyArea([key_1])
            if area.descriptor_decimal in available[0]:
                add_choice(area.value, value_1.get('label'), 0)
            for key_2, value_2 in value_1.get('children', {}).items():
                area = PolicyArea([key_1, key_2])
                if area.descriptor_decimal in available[1]:
                    add_choice(area.value, value_2.get('label'), 1)
                for key_3, value_3 in value_2.get('children', {}).items():
                    area = PolicyArea([key_1, key_2, key_3])
                    if area.descriptor_decimal in available[2]:
                        add_choice(area.value, value_3.get('label'), 2)

    def on_request(self):
        self.delete_field('csrf_token')
        self.populate_policy_area()

    def apply_model(self, model):
        self.policy_area.data = model.policy_area
        self.legal_form.data = model.legal_form
        self.result.data = model.result
        self.from_date.data = model.from_date
        self.to_date.data = model.to_date
        self.sort_by.data = model.sort_by
        self.sort_order.data = model.sort_order

        # default unselected checkboxes to all choices
        if not self.legal_form.data:
            self.legal_form.data = list(LEGAL_FORM.keys())
        if not self.result.data:
            self.result.data = list(RESULT.keys())



from onegov.core.security import Private
from onegov.core.security import Public
from onegov.form import Form
from onegov.swissvotes import _
from onegov.swissvotes import SwissvotesApp
from onegov.swissvotes.forms import AttachmentsForm
from onegov.swissvotes.layouts import DeleteVoteLayout
from onegov.swissvotes.layouts import UploadVoteAttachemtsLayout
from onegov.swissvotes.layouts import VoteLayout
from onegov.swissvotes.models import SwissVote


@SwissvotesApp.html(
    model=SwissVote,
    permission=Public,
    template='vote.pt'
)
def view_vote(self, request):
    return {
        'layout': VoteLayout(self, request)
    }


@SwissvotesApp.form(
    model=SwissVote,
    permission=Private,
    form=AttachmentsForm,
    template='form.pt',
    name='upload'
)
def upload_vote_attachments(self, request, form):
    layout = UploadVoteAttachemtsLayout(self, request)

    if form.submitted(request):
        form.update_model(self)
        request.message(_("Attachments updated"), 'success')
        return request.redirect(request.link(self))

    if not form.errors:
        form.apply_model(self)

    return {
        'layout': layout,
        'form': form,
        'cancel': request.link(self)
    }


@SwissvotesApp.form(
    model=SwissVote,
    permission=Private,
    form=Form,
    template='form.pt',
    name='delete'
)
def delete_vote(self, request, form):
    layout = DeleteVoteLayout(self, request)

    if form.submitted(request):
        request.session.delete(self)
        request.message(_("Vote deleted"), 'success')
        return request.redirect(layout.votes_link)

    return {
        'layout': layout,
        'form': form,
        'subtitle': self.title,
        'message': _(
            'Do you really want to delete "${item}"?',
            mapping={'item': self.title}
        ),
        'button_text': _("Delete vote"),
        'button_class': 'alert',
        'cancel': request.link(self)
    }



import logging

from translationstring import TranslationStringFactory

log = logging.getLogger('onegov.swissvotes')
log.addHandler(logging.NullHandler())
_ = TranslationStringFactory('onegov.swissvotes')

from onegov.swissvotes.app import SwissvotesApp  # noqa

__all__ = (
    '_',
    'log',
    'SwissvotesApp'
)



from wtforms import SelectMultipleField
from wtforms.widgets import Select
from onegov.swissvotes.models import PolicyArea


class PolicyAreaWidget(Select):
    """ The widget for the chosen field.

    Adds custom classes based on provided value.

    """

    @classmethod
    def render_option(cls, value, label, selected, **kwargs):
        kwargs['class'] = 'level-{}'.format(PolicyArea(value).level)
        return super(PolicyAreaWidget, cls).render_option(
            value, label, selected, **kwargs
        )


class PolicyAreaField(SelectMultipleField):
    """ A select field with chosen support. """

    widget = PolicyAreaWidget(multiple=True)

    def __init__(self, *args, **kwargs):
        render_kw = kwargs.pop('render_kw', {})
        render_kw['class_'] = 'chosen-select'
        kwargs['render_kw'] = render_kw

        super().__init__(*args, **kwargs)



from morepath.request import Response
from onegov.core.security import Private
from onegov.core.security import Public
from onegov.core.security import Secret
from onegov.form import Form
from onegov.swissvotes import _
from onegov.swissvotes import SwissvotesApp
from onegov.swissvotes.collections import SwissVoteCollection
from onegov.swissvotes.forms import SearchForm
from onegov.swissvotes.forms import UpdateDatasetForm
from onegov.swissvotes.layouts import DeleteVotesLayout
from onegov.swissvotes.layouts import UpdateVotesLayout
from onegov.swissvotes.layouts import VotesLayout
from translationstring import TranslationString


@SwissvotesApp.form(
    model=SwissVoteCollection,
    permission=Public,
    form=SearchForm,
    template='votes.pt'
)
def view_votes(self, request, form):
    if not form.errors:
        form.apply_model(self)

    return {
        'layout': VotesLayout(self, request),
        'form': form
    }


@SwissvotesApp.form(
    model=SwissVoteCollection,
    permission=Private,
    form=UpdateDatasetForm,
    template='form.pt',
    name='update'
)
def update_votes(self, request, form):
    layout = UpdateVotesLayout(self, request)

    if form.submitted(request):
        self.update(form.dataset.data)
        request.message(_("Dataset updated"), 'success')

        # Warn if descriptor labels are missing
        missing = set()
        for vote in self.query():
            for policy_area in vote.policy_areas:
                missing |= set(
                    path for path in policy_area.label_path
                    if not isinstance(path, TranslationString)
                )
        if missing:
            request.message(
                _(
                    "The dataset contains unknown descriptors: ${items}.",
                    mapping={'items': ', '.join(sorted(missing))}
                ),
                'warning'
            )

        return request.redirect(layout.votes_link)

    return {
        'layout': layout,
        'form': form,
        'cancel': request.link(self)
    }


@SwissvotesApp.view(
    model=SwissVoteCollection,
    permission=Public,
    name='csv'
)
def export_votes(self, request):
    response = Response(
        content_type='text/csv',
        content_disposition='inline; filename=dataset.csv'
    )
    self.export(response.body_file)
    return response


@SwissvotesApp.form(
    model=SwissVoteCollection,
    permission=Secret,
    form=Form,
    template='form.pt',
    name='delete'
)
def delete_vote(self, request, form):
    layout = DeleteVotesLayout(self, request)

    if form.submitted(request):
        for vote in self.query():
            request.session.delete(vote)
        request.message(_("All votes deleted"), 'success')
        return request.redirect(layout.votes_link)

    return {
        'layout': layout,
        'form': form,
        'message': _("Do you really want to delete all votes?!"),
        'button_text': _("Delete all votes"),
        'button_class': 'alert',
        'cancel': request.link(self)
    }



from onegov.swissvotes.models.page import TranslatablePage
from onegov.swissvotes.models.policy_area import PolicyArea
from onegov.swissvotes.models.principal import Principal
from onegov.swissvotes.models.vote import SwissVote
from onegov.swissvotes.models.vote import SwissVoteFile


__all__ = (
    'PolicyArea',
    'Principal',
    'SwissVote',
    'SwissVoteFile',
    'TranslatablePage',
)



from onegov.swissvotes.fields.csv import SimpleCsvUploadField
from onegov.swissvotes.fields.dataset import SwissvoteDatasetField
from onegov.swissvotes.fields.policy_area import PolicyAreaField


__all__ = (
    'PolicyAreaField',
    'SimpleCsvUploadField',
    'SwissvoteDatasetField',
)
